from typing import Sequence

def attach_and_exec(pid: int, python_code: str, /) -> None: ...
def exec_in_threads(pid: int, tids: Sequence[int], python_code: str, /) -> None: ...
def reap_process(pid: int, timeout_ms: int) -> int: ...

class AttachError(Exception): ...
class UnknownStateError(AttachError): ...
class InterruptedError(AttachError): ...
class ReapError(Exception): ...
class SignalledError(ReapError): ...



from typing import Any, Type, Union

from pymontrace.tracee import Quantization

def create(fd: int) -> Any: ...
def create_agg_buffer(fd: int, name: str) -> Any: ...
def open_agg_buffer(fd: int) -> Any: ...
def encode_entry(key_data: bytes, value, Quantization: Type) -> bytes: ...
def encode_value(value, Quantization: Type) -> bytes: ...
def decode_value(data, Quantization: Type) -> Union[int, float, Quantization]: ...



import os
import pickle
import threading
from typing import Union

from pymontrace import _tracebuffer

__all__ = [
    'create', 'create_agg_buffer', 'open_agg_buffer', 'encode_entry',
    'encode_value', 'decode_value'
]


PAGESIZE = os.sysconf("SC_PAGE_SIZE")
DEFAULT_BUFFER_SIZE = (1 << 20)  # 1MiB


class TraceBuffer:
    def __init__(self, _tb) -> None:
        self._tb = _tb
        self._lock = threading.Lock()

    def close(self):
        self._tb = None  # __del__ will clean up

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def read(self) -> bytes:
        if (tb := self._tb) is None:
            raise ValueError('I/O operation on closed buffer')
        return tb.read()

    def write(self, data: bytes) -> None:
        if (tb := self._tb) is None:
            raise ValueError('I/O operation on closed buffer')

        if not self._lock.acquire(timeout=1.0):
            raise TimeoutError('failed to acquire lock on TraceBuffer after 1s')
        try:
            return tb.write(data)
        finally:
            self._lock.release()


def create(filename: str, size: int = DEFAULT_BUFFER_SIZE) -> TraceBuffer:
    if size < PAGESIZE or size % PAGESIZE != 0:
        raise ValueError("Invalid size, must a multiple of PAGESIZE")
    fd = os.open(filename, os.O_CREAT | os.O_RDWR)
    os.ftruncate(fd, size)
    with open(fd, 'a+b', buffering=0) as f:  # <- closes the fd
        tb = _tracebuffer.create(f.fileno())
    return TraceBuffer(tb)


class AggBuffer:
    def __init__(self, _agg_buffer, fn: str) -> None:
        self._agg_buffer = _agg_buffer
        self.fn = fn
        self._lock = threading.Lock()
        self.epoch = 2

    @property
    def name(self) -> str:  # tracer
        return self._agg_buffer.name

    @property
    def agg_op(self):  # tracer
        return self._agg_buffer.agg_op

    @agg_op.setter
    def agg_op(self, value: int):  # tracee
        self._agg_buffer.agg_op = value

    def __enter__(self):
        if not self._lock.acquire(timeout=1.0):
            raise TimeoutError('failed to acquire lock on AggBuffer after 1s')
        self.epoch = self._agg_buffer.epoch

    def __exit__(self, exc_type, exc_value, traceback):
        self._lock.release()

    def read(self, offset: int, size: int) -> bytes:  # tracee
        return self._agg_buffer.read(self.epoch, offset, size)

    def write(self, kvp: bytes) -> tuple[int, int]:  # tracee
        return self._agg_buffer.write(self.epoch, kvp)

    def update(self, kvp: bytes, offset: int, size: int) -> bytes:  # tracee
        return self._agg_buffer.update(self.epoch, kvp, offset, size)

    def readall(self, epoch: Union[int, None] = None) -> bytes:  # tracer
        if epoch is None:
            epoch = self.epoch
        return self._agg_buffer.readall(epoch)

    def written(self, epoch: int) -> int:  # tracer
        """The number of bytes already written to epoch buffer"""
        return self._agg_buffer.written(epoch)

    def switch(self):  # tracer
        """Switch the active buffer and increment the epoch"""
        # Maybe this should just be combined into a single C function?
        self._agg_buffer.reset(self.epoch - 1)
        self.epoch = self._agg_buffer.incr_epoch()


def create_agg_buffer(name: str, filename: str, size=DEFAULT_BUFFER_SIZE) -> AggBuffer:
    with open(filename, 'x+b', buffering=0) as f:
        f.truncate(size)
        ab = _tracebuffer.create_agg_buffer(f.fileno(), name)
    return AggBuffer(ab, filename)


def open_agg_buffer(filename: str, size=DEFAULT_BUFFER_SIZE) -> AggBuffer:
    with open(filename, 'r+b', buffering=0) as f:
        f.truncate(size)
        ab = _tracebuffer.open_agg_buffer(f.fileno())
    return AggBuffer(ab, filename)


def encode_entry(key, value, Quantization) -> bytes:
    key_data = pickle.dumps(key)
    return _tracebuffer.encode_entry(key_data, value, Quantization)


# We skip the wrapper function as it adds about 40ns overhead and this is used
# in the hot path.
encode_value = _tracebuffer.encode_value

decode_value = _tracebuffer.decode_value



import argparse
import atexit
import enum
import os
import socket
import subprocess
import sys

import pymontrace.attacher
from pymontrace import tracer
from pymontrace.tracer import (
    CommsFile, TraceState, convert_probe_filter, create_and_bind_socket,
    decode_and_print_forever, decode_and_print_remaining, encode_script,
    format_bootstrap_snippet, format_untrace_snippet, to_remote_path,
    validate_script
)


# Getting the version from package metadata adds quite a bit of overhead
# so we only do it if requested.
class VersionAction(argparse.Action):
    def __init__(self, option_strings, dest, **kwargs):
        help = "print pymontrace's version"
        super().__init__(option_strings, dest, nargs=0, help=help)

    def _get_version(self):
        import importlib.metadata
        return importlib.metadata.version("pymontrace")

    def __call__(self, parser, namespace, values, option_string=None):
        print(self._get_version())
        sys.exit(0)


parser = argparse.ArgumentParser(
    prog='pymontrace',
    description=(
        'Attaches to a running python program, or starts one, and injects '
        'debugging statements into selected probe sites.'
    ),
    allow_abbrev=False,
)

parser.add_argument('--version', action=VersionAction)
target_group = parser.add_argument_group('target selection')
target_group_alts = target_group.add_mutually_exclusive_group(required=True)
target_group_alts.add_argument(
    '-c', dest='pyprog',
    help=(
        "a python script to run and trace, including any arguments "
        "e.g. 'some_script.py \"one arg\" another'"
    ),
)
target_group_alts.add_argument(
    '-p', dest='pid', type=int,
    help='pid of a python process to attach to',
)
# used internally for handling -c',
target_group_alts.add_argument(
    '-X', dest='subproc',
    help=argparse.SUPPRESS,
)

action_group = parser.add_argument_group('action')
action_group_alts = action_group.add_mutually_exclusive_group(required=True)
action_group_alts.add_argument(
    '-e', dest='prog_text', type=str,
    help="pymontrace program text e.g. 'line:*script.py:13 {{ print(ctx.a, ctx.b) }}'",
)
action_group_alts.add_argument(
    '-l', dest='probe_filter', type=str,
    help=(
        "list available probe sites, takes a probe spec as a filter "
        "e.g. 'func:threading.*:start'"
    )
)
# If/when we allow script files
#
# action_group_alts.add_argument(
#     'script',
#     type=argparse.FileType('r'),
#     nargs='?',
#     help='a pymontrace script file'
# )


def force_unlink(path):
    try:
        os.unlink(path)
    except Exception:
        pass


class EndReason(enum.Enum):
    DISCONNECTED = enum.auto()
    EXITED = enum.auto()
    ENDED_EARLY = enum.auto()
    INTERRUPTED = enum.auto()


def receive_and_print_until_interrupted(
    pid: int, s: socket.socket, tracestate: TraceState
) -> EndReason:
    print('Probes installed. Hit CTRL-C to end...', file=sys.stderr)
    try:
        outcome = decode_and_print_forever(pid, s, tracestate)
        if outcome == tracer.DecodeEndReason.DISCONNECTED:
            print('Target disconnected.', file=sys.stderr)
            return EndReason.DISCONNECTED
        elif outcome == tracer.DecodeEndReason.EXITED:
            return EndReason.EXITED
        elif outcome == tracer.DecodeEndReason.ENDED_EARLY:
            return EndReason.ENDED_EARLY
        else:
            tracer.assert_never(outcome)
    except KeyboardInterrupt:
        return EndReason.INTERRUPTED


class PIDState(enum.Enum):
    GONE = enum.auto()
    STILL_THERE = enum.auto()


def wait_till_gone(pid: int, timeout=1.0) -> PIDState:
    try:
        exitstatus = pymontrace.attacher.reap_process(pid, int(timeout * 1000))
        print(f'Target exited with status {exitstatus}.', file=sys.stderr)
        return PIDState.GONE
    except TimeoutError:
        os.kill(pid, 0)  # A kind of assert
        return PIDState.STILL_THERE
    except ProcessLookupError:
        return PIDState.GONE
    except pymontrace.attacher.SignalledError as e:
        signum, desc = e.args
        print(f'Target killed by signal: {desc} ({signum})')
        return PIDState.GONE


def tracepid(pid: int, encoded_script: bytes):
    try:
        os.kill(pid, 0)
    except ProcessLookupError:
        parser.error(f"no such process: {pid}")

    tracer.install_signal_handler()

    site_extension = tracer.install_pymontrace(pid)

    comms = CommsFile(pid)
    atexit.register(force_unlink, comms.localpath)

    with create_and_bind_socket(comms, pid) as ss:
        # requires sudo on mac
        pymontrace.attacher.attach_and_exec(
            pid,
            format_bootstrap_snippet(
                encoded_script, comms.remotepath,
                to_remote_path(pid, site_extension.name),
            ),
        )

        ss.settimeout(1.0)
        s, _ = ss.accept()
        # TODO: verify the connected party is pid
        os.unlink(comms.localpath)

        tracestate = TraceState()

        outcome = receive_and_print_until_interrupted(pid, s, tracestate)
        if (outcome == EndReason.INTERRUPTED
                or wait_till_gone(pid) == PIDState.STILL_THERE):
            print('Removing probes...', file=sys.stderr)
            # BUG: this may hang if end probes send too much data
            pymontrace.attacher.attach_and_exec(
                pid,
                format_untrace_snippet(),
            )
        decode_and_print_remaining(pid, s, tracestate)


def subprocess_entry(progpath, encoded_script: bytes):
    import runpy
    import shlex
    import time

    from pymontrace.tracee import connect, remote, settrace, unsettrace

    sys.argv = shlex.split(progpath)

    comm_file = CommsFile(os.getpid()).remotepath
    while not os.path.exists(comm_file):
        time.sleep(0.1)
    connect(comm_file)

    # Avoid code between settrace and starting the target program
    settrace(encoded_script)
    try:
        runpy.run_path(sys.argv[0], run_name='__main__')
    except KeyboardInterrupt:
        pass
    finally:
        unsettrace(preclose=remote.notify_exit)


def tracesubprocess(progpath: str, prog_text):
    p = subprocess.Popen(
        [sys.executable, '-m', 'pymontrace', '-X', progpath, '-e', prog_text]
    )

    comms = CommsFile(p.pid)
    atexit.register(force_unlink, comms.localpath)

    with create_and_bind_socket(comms, p.pid) as ss:
        s, _ = ss.accept()
        os.unlink(comms.localpath)

        tracestate = TraceState()
        outcome = receive_and_print_until_interrupted(p.pid, s, tracestate)
        # The child will also have had a SIGINT at this point as it's
        # in the same terminal group. So should have ended unless it's
        # installed its own signal handlers.
        decode_and_print_remaining(p.pid, s, tracestate)

        if outcome == EndReason.ENDED_EARLY:
            # User included an [pmt.]exit() call in their script
            # We own the subprocess - if the tracing stops, we kill this
            # patient to avoid an orphan.
            p.terminate()


def cli_main():
    args = parser.parse_args()

    if args.probe_filter:
        try:
            args.prog_text = convert_probe_filter(args.probe_filter)
        except Exception as e:
            parser.error(str(e))

    try:
        validate_script(args.prog_text)
    except Exception as e:
        parser.error(str(e))

    if args.pyprog:
        tracesubprocess(args.pyprog, args.prog_text)
    elif args.subproc:
        subprocess_entry(args.subproc, encode_script(args.prog_text))
    elif args.pid:
        tracepid(args.pid, encode_script(args.prog_text))
    else:
        parser.error("one of -p or -c required")


if __name__ == '__main__':
    cli_main()



import enum
import inspect
import io
import operator
import os
import pathlib
import re
import selectors
import shutil
import signal
import socket
import struct
import sys
import textwrap
import threading
import traceback
from contextlib import contextmanager
from dataclasses import dataclass
from tempfile import TemporaryDirectory
from typing import NoReturn, Union

from pymontrace import _darwin, attacher, tracebuffer
from pymontrace.tracebuffer import AggBuffer, TraceBuffer
from pymontrace.tracee import PROBES_BY_NAME, Quantization


# Replace with typing.assert_never after 3.11
def assert_never(arg: NoReturn) -> NoReturn:
    raise AssertionError(f"assert_never: got {arg!r}")


def parse_script(script_text: str):
    # In here we use i for "input"

    class ParseError(Exception):
        def __init__(self, message) -> None:
            super().__init__(message)
            self.message = message

    class ExpectError(ParseError):
        def __init__(self, expected, got=None) -> None:
            super().__init__(f'Expected {expected} got {got}')

    def expect(s):
        def f(i):
            if i.startswith(s):
                return s, i[len(s):]
            got = i if len(i) <= len(s) else f'{i[:len(s)]}...'
            raise ExpectError(s, got)
        return f

    def regex_parser(regex, desc=None):
        outer_desc = desc

        def parse_regex(i: str, desc=None):
            if m := re.match(regex, i):
                return m[0], i[len(m[0]):]
            assert desc is not None or outer_desc is not None, "regex_parser no desc"
            raise ExpectError(desc or outer_desc, got=i)
        return parse_regex

    def many(pred):
        def parsemany(i):
            c = 0
            while i[c:] and pred(i[c]):
                c += 1
            return i[:c], i[c:]
        return parsemany

    def manyone(pred, desc):
        thismany = many(pred)

        def parsemanyone(i):
            if not pred(i[:1]):
                raise ExpectError(desc, got=i)
            return thismany(i)
        return parsemanyone

    def whilenot(literal):
        def parseuntil(i):
            j = i
            c = 0
            while j and not j.startswith(literal):
                j = j[1:]
                c += 1
            if j.startswith(literal):
                return i[:c], i[c:]
            raise ParseError('Ending {literal!r} not found, got {i!r}')
        return parseuntil

    whitespace = many(str.isspace)
    nonempty_whitespace = manyone(str.isspace, 'whitespace')  # noqa

    parse_colon = expect(':')
    parse_probe_name = regex_parser(r'[^:\s]+', 'probe name')
    parse_arg1 = regex_parser(r'[^:\s]*')
    parse_arg2 = regex_parser(r'[^:\s{]+')

    def parse_probe_spec(i):
        name, i = parse_probe_name(i)
        valid_probe_names = ('line', 'pymontrace', 'func')
        if name not in valid_probe_names:
            raise ParseError(
                f'Unknown probe {name!r}. '
                f'Valid probes are: {", ".join(valid_probe_names)}'
            )
        arg1_desc = {
            'line': 'file path',
            'pymontrace': 'nothing',
            'func': 'qualified function path (qpath)'
        }[name]
        arg2_desc = {
            'line': 'line number',
            'pymontrace': 'BEGIN or END',
            'func': 'func probe point'
        }[name]

        _, i = parse_colon(i)
        arg1, i = parse_arg1(i, arg1_desc)
        _, i = parse_colon(i)
        arg2, i = parse_arg2(i, arg2_desc)
        if name == 'line':
            _ = int(arg2)  # just validate
        elif name == 'pymontrace':
            if arg2 not in ('BEGIN', 'END'):
                raise ParseError(
                    f'Invalid probe point for pymontrace: {arg2}. '
                    'Valid pymontrace probe specs are: '
                    'pymontrace::BEGIN, pymontrace::END'
                )
        elif name == 'func':
            if any(not (c.isalnum() or c in '*._') for c in arg1):
                raise ParseError(f'Invalid qpath glob: {arg1!r}')
            if arg2 not in ('start', 'yield', 'resume', 'return', 'unwind'):
                raise ParseError(f'Invalid func probe point {arg2!r}')
        else:
            assert_never(name)
        return (name, arg1, arg2), i

    parse_action_start = expect('{{')
    parse_action_body = whilenot('}}')
    parse_action_end = expect('}}')

    def parse_probe_action(i: str):
        _, i = parse_action_start(i)
        inner, i = parse_action_body(i)
        _, i = parse_action_end(i)
        return inner, i

    probe_actions: list[tuple[tuple[str, str, str], str]] = []

    i = script_text
    _, i = whitespace(i)  # eat leading space
    while i:
        probespec, i = parse_probe_spec(i)
        _, i = whitespace(i)
        action, i = parse_probe_action(i)

        # Should we check it's valid python? this may not be the target's
        # python...
        action = textwrap.dedent(action)
        compile(action, '<probeaction>', 'exec')

        probe_actions.append((probespec, action))
        _, i = whitespace(i)
    return probe_actions


def validate_script(script_text: str) -> str:
    """
    Raises an exception if the script text is invalid.
    Returns the script text if it's valid.
    """
    _ = parse_script(script_text)
    return script_text


def convert_probe_filter(probe_filter: str) -> str:

    # This really ought to reuse the parsing logic above...

    parts = probe_filter.strip().split(':')

    if len(parts) > 3:
        raise ValueError('Too many probe parts. Expected at most 3')
    if len(parts) >= 3:
        # want to catch accidental mixup of -e and -l
        final_part = parts[2]
        if '{' in final_part:
            pf = probe_filter.strip()
            idx = pf.find('{', pf.find(':', pf.find(':') + 1) + 1)
            additional = pf + '\n' + (' ' * idx) + '^~'
            raise ValueError(
                "Unexpected '{' in probe filter. "
                "Did you mean to use the -e option instead of -l?"
                f"\n{additional}"
            )
    if ' ' in (pf := probe_filter.strip()):
        idx = pf.find(' ')
        additional = pf + '\n' + (' ' * idx) + '^~'
        raise ValueError(f"Unexpected space in probe filter.\n{additional}")

    if parts[0] not in ('func', 'line', 'pymontrace'):
        pf = probe_filter.strip()
        additional = pf + '\n' + '^' + ('~' * (len(parts[0]) - 1))
        raise ValueError(
            f'Unknown probe name {parts[0]!r}. '
            'Expected one of func, line, pymontrace.'
            f"\n{additional}"
        )

    args = ', '.join(f'{part!r}' for part in parts)
    # As two separate probes so the exit doesn't fail to happen on exception
    prog_text = 'pymontrace::BEGIN {{ printprobes(' + args + ') }} ' \
        + 'pymontrace::BEGIN {{ exit() }}'
    return prog_text


def _encode_script(parsed_script) -> bytes:

    VERSION = 1
    result = bytearray()
    num_probes = len(parsed_script)
    result += struct.pack('=HH', VERSION, num_probes)

    for (probe_spec, action) in parsed_script:
        name, *args = probe_spec
        probe_id = PROBES_BY_NAME[name].id
        result += struct.pack('=BB', probe_id, len(args))
        for arg in args:
            result += arg.encode()
            result += b'\x00'
        result += action.encode()
        result += b'\x00'
    return bytes(result)


def encode_script(script_text: str) -> bytes:
    parsed = parse_script(script_text)
    return _encode_script(parsed)


def install_pymontrace(pid: int) -> TemporaryDirectory:
    """
    In order that pymontrace can be used without prior installatation
    we prepare a module containing the tracee parts and extends
    """
    import pymontrace
    import pymontrace._tracebuffer
    import pymontrace.tracebuffer
    import pymontrace.tracee

    # Maybe there will be cases where checking for some TMPDIR is better.
    # but this seems to work so far.
    ptmpdir = '/tmp'
    if sys.platform == 'linux' and os.path.isdir(f'/proc/{pid}/root/tmp'):
        ptmpdir = f'/proc/{pid}/root/tmp'

    tmpdir = TemporaryDirectory(dir=ptmpdir)
    # Would be nice to change this so the owner group is the target gid
    os.chmod(tmpdir.name, 0o755)
    moddir = pathlib.Path(tmpdir.name) / 'pymontrace'
    moddir.mkdir()

    for module in [pymontrace, pymontrace.tracee, pymontrace.tracebuffer]:
        source_file = inspect.getsourcefile(module)
        if source_file is None:
            raise FileNotFoundError('failed to get source for module', module)

        shutil.copyfile(source_file, moddir / os.path.basename(source_file))

    for module in [pymontrace._tracebuffer]:
        if module.__spec__ is None:
            raise RuntimeError(f'{module.__name__} missing __spec__ attribute')
        shared_object = module.__spec__.origin
        if shared_object is None:
            raise FileNotFoundError('failed to find shared object for module', module)
        shutil.copyfile(shared_object, moddir / os.path.basename(shared_object))

    return tmpdir


def to_remote_path(pid: int, path: str) -> str:
    proc_root = f'/proc/{pid}/root'
    if path.startswith(f'{proc_root}/'):
        return path[len(proc_root):]
    return path


def from_remote_path(pid: int, remote_path: str) -> str:
    """
    Converts a path that makes sense for the tracee to one that represents
    the same file from the perspective of the tracer
    """
    assert remote_path[0] == '/'
    # Trailing slash needed otherwise it's the symbolic link
    pidroot = f'/proc/{pid}/root/'
    if (os.path.isdir(pidroot) and not os.path.samefile(pidroot, '/')):
        return f'{pidroot}{remote_path[1:]}'
    else:
        return remote_path


def format_bootstrap_snippet(encoded_script: bytes, comm_file: str, site_extension: str):

    # Running settrace in a nested function does two things
    #  1. it keeps the locals dictionary clean in the injection site
    #  2. it fixes a bug that means that the top level of an interactive
    #  session could not be traced.
    return textwrap.dedent(
        f"""
        def _pymontrace_bootstrap():
            import sys
            do_unload = 'pymontrace.tracee' not in sys.modules
            try:
                import pymontrace.tracee
                pymontrace.tracee.do_unload = do_unload
            except Exception:
                sys.path.append('{site_extension}')
                try:
                    import pymontrace.tracee
                    pymontrace.tracee.do_unload = do_unload
                finally:
                    sys.path.remove('{site_extension}')
            pymontrace.tracee.connect({comm_file!r})
            pymontrace.tracee.settrace({encoded_script!r})
        _pymontrace_bootstrap()
        del _pymontrace_bootstrap
        """
    )


def format_additional_thread_snippet():
    return textwrap.dedent(
        """
        try:
            import pymontrace.tracee
            pymontrace.tracee.synctrace()
            del pymontrace
        except Exception:
            pass
        """
    )


def format_untrace_snippet():
    return textwrap.dedent(
        """
        import pymontrace.tracee
        pymontrace.tracee.unsettrace()
        if getattr(pymontrace.tracee, 'do_unload', False):
            del __import__('sys').modules['pymontrace.tracee']
            del __import__('sys').modules['pymontrace.tracebuffer']
            del __import__('sys').modules['pymontrace._tracebuffer']
            del __import__('sys').modules['pymontrace']
        del pymontrace
        """
    )


class CommsFile:
    """
    Defines where the communication socket is bound. Primarily for Linux,
    where the target may have another root directory, we define `remotepath`
    for use inside the tracee, once attached. `localpath` is where the tracer
    will create the socket in it's own view of the filesystem.
    """
    def __init__(self, pid: int):
        # TODO: We should probably add a random component with mktemp...
        self.remotepath = f'/tmp/pymontrace-{pid}'
        self.localpath = from_remote_path(pid, self.remotepath)


def get_proc_euid(pid: int):
    if sys.platform == 'darwin':
        # A subprocess alternative would be:
        #   ps -o uid= -p PID
        return _darwin.get_euid(_darwin.kern_proc_info(pid))
    if sys.platform == 'linux':
        # Will this work if it's in a container ??
        with open(f'/proc/{pid}/status') as f:
            for line in f:
                if line.startswith('Uid:'):
                    # Linux: fs/proc/array.c (or
                    #        Documentation/filesystems/proc.rst)
                    # Uid:	uid	euid	suid	fsuid
                    return int(line.split('\t')[2])
            return None
    raise NotImplementedError


def is_own_process(pid: int):
    # euid is the one used to decide on access permissions.
    return get_proc_euid(pid) == os.geteuid()


@contextmanager
def set_umask(target_pid: int):
    # A future idea could be to get the gid of the target
    # and give their group group ownership.
    if not is_own_process(target_pid):
        saved_umask = os.umask(0o000)
        try:
            yield
        finally:
            os.umask(saved_umask)
    else:
        yield


def create_and_bind_socket(comms: CommsFile, pid: int) -> socket.socket:
    ss = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    with set_umask(pid):
        ss.bind(comms.localpath)
    ss.listen(0)
    return ss


def get_peer_pid(s: socket.socket):
    if sys.platform == 'darwin':
        # See: sys/un.h
        SOL_LOCAL = 0
        LOCAL_PEERPID = 0x002
        peer_pid_buf = s.getsockopt(SOL_LOCAL, LOCAL_PEERPID, 4)
        return int.from_bytes(peer_pid_buf, sys.byteorder)
    if sys.platform == 'linux':
        ucred_buf = s.getsockopt(socket.SOL_SOCKET, socket.SO_PEERCRED, 12)
        (pid, uid, gid) = struct.unpack('iii', ucred_buf)
        return pid
    raise NotImplementedError


def settrace_in_threads(pid: int, thread_ids: 'tuple[int]'):
    try:
        attacher.exec_in_threads(
            pid, thread_ids, format_additional_thread_snippet()
        )
    except NotImplementedError:
        print(
            f'There are an additional {len(thread_ids)} threads '
            'that are not able to be traced', sys.stderr,
        )


signal_read, signal_write = socket.socketpair()


def signal_handler(signo: int, frame):
    try:
        signal_write.send(signo.to_bytes(1, sys.byteorder))
        # We close the write end of the pair so that we can exit it the
        # decode and print loop hangs due to a misbehaving tracee and the user
        # or OS sends a second "I'm impatient" signal
        signal_write.close()
    except OSError:
        # We implement the default behaviour, i.e. terminating. But
        # we raise SystemExit so that finally blocks are run and atexit.
        raise SystemExit(128 + signo)


def install_signal_handler():
    for signo in [
            signal.SIGINT,
            signal.SIGHUP,
            signal.SIGTERM,
            signal.SIGQUIT
    ]:
        signal.signal(signo, signal_handler)


class WaitResult(enum.Enum):
    READY = enum.auto()
    TIMEOUT = enum.auto()


def wait_till_ready_or_got_signal(
    s: socket.socket, sel: selectors.BaseSelector
) -> WaitResult:
    ready = sel.select(timeout=0.1)
    if len(ready) == 0:
        return WaitResult.TIMEOUT
    for key, _ in ready:
        if key.fileobj == signal_read:
            received = signal_read.recv(1)
            if received == b'':
                sel.unregister(signal_read)
                continue
            signo = int.from_bytes(received, sys.byteorder)
            if signo == signal.SIGINT:
                raise KeyboardInterrupt
            else:
                raise SystemExit(128 + signo)

        assert key.fileobj == s
    # If we make it out of that for loop it means s is ready
    return WaitResult.READY


class TraceState:
    def __init__(self):
        self.trace_buffers: list[TraceBuffer] = []
        self.agg_buffers: list[AggBuffer] = []

        # To not exhaust the tracee's memory,
        # periodically aggregation data is accumulated here in the tracer.
        self.chkptd_aggs = {}


def decode_trace_buffers_once(buffers: list[TraceBuffer]):
    from pymontrace.tracee import Message

    header_fmt = struct.Struct('=HH')

    for buf in buffers:
        data = buf.read()
        while data != b'':
            (kind, size) = header_fmt.unpack_from(data)
            offset = header_fmt.size
            body = data[offset:offset + size]
            data = data[offset + size:]
            if kind in (Message.PRINT, Message.ERROR,):
                line = body
                out = (sys.stderr if kind == Message.ERROR else sys.stdout)
                out.write(line.decode())
            else:
                print(f'unexpected data in trace buffer: {kind=}', file=sys.stderr)


def accumulate_func(agg_op: int):
    from pymontrace.tracee import AggOp

    if agg_op == AggOp.NONE:
        return lambda _, v2: v2
    elif agg_op == AggOp.COUNT:
        return operator.add
    elif agg_op == AggOp.SUM:
        return operator.add
    elif agg_op == AggOp.MAX:
        return max
    elif agg_op == AggOp.MIN:
        return min
    elif agg_op == AggOp.QUANTIZE:
        return operator.add
    raise ValueError(f'invalid agg_op: {agg_op}')


def read_records(agg_buffer: AggBuffer, epoch: Union[int, None] = None):

    f = io.BytesIO(agg_buffer.readall(epoch))

    while True:
        key_len_bytes = f.read(4)
        if len(key_len_bytes) < 4:
            break
        length = int.from_bytes(key_len_bytes, sys.byteorder)
        if length == 0:
            break
        key_bytes = f.read(length)
        if len(key_bytes) < length:
            break

        value_len_bytes = f.read(4)

        if len(value_len_bytes) < 4:
            break
        value_len = int.from_bytes(value_len_bytes, sys.byteorder)
        if value_len == 0:
            break
        value_bytes = f.read(value_len)
        if len(value_bytes) < value_len:
            break

        yield key_len_bytes + key_bytes + value_len_bytes + value_bytes


def accumulate_buffer(buffer: AggBuffer, epoch: int, ts: TraceState):
    from pymontrace.tracee import PMTMap

    accumulated = ts.chkptd_aggs.setdefault(buffer.name, {})

    agg_op = buffer.agg_op
    accumulate = accumulate_func(agg_op)

    for record_bytes in read_records(buffer, epoch=epoch):
        key, value = PMTMap._decode(record_bytes)
        if key in accumulated:
            existing = accumulated[key]
            accumulated[key] = accumulate(existing, value)
        else:
            accumulated[key] = value


def switch_aggregation_buffers(ts: TraceState):

    for buffer in ts.agg_buffers:

        # First check if the active buffer has been written to
        # (otherwise it's not really active).
        active = buffer.epoch
        if buffer.written(active) == 0:
            continue

        which = buffer.epoch - 1
        accumulate_buffer(buffer, which, ts)
        buffer.switch()


class DecodeEndReason(enum.Enum):
    DISCONNECTED = enum.auto()
    EXITED = enum.auto()
    ENDED_EARLY = enum.auto()


def decode_and_print_forever(
    pid: int, s: socket.socket, ts: TraceState, *, only_print=False
):
    from pymontrace.tracee import Message
    EVENT_READ = selectors.EVENT_READ

    sel = selectors.DefaultSelector()
    sel.register(signal_read, EVENT_READ)
    sel.register(s, EVENT_READ)

    t = None
    try:
        header_fmt = struct.Struct('=HH')
        while True:
            # We check for signals between message receipt so that s remains
            # in a good state to read final shutdown messages (e.g. from
            # the pymontrace::END probe)
            if wait_till_ready_or_got_signal(s, sel) == WaitResult.TIMEOUT:
                decode_trace_buffers_once(ts.trace_buffers)
                switch_aggregation_buffers(ts)
                continue

            header = s.recv(header_fmt.size)
            if header == b'':
                return DecodeEndReason.DISCONNECTED
            (kind, size) = header_fmt.unpack(header)
            body = s.recv(size)

            if kind == Message.THREADS and only_print:
                print(f'ignoring {kind=} during shutdown', file=sys.stderr)
            elif kind == Message.THREADS:
                count_threads = size // struct.calcsize('=Q')
                thread_ids = struct.unpack('=' + (count_threads * 'Q'), body)
                t = threading.Thread(target=settrace_in_threads,
                                     args=(pid, thread_ids), daemon=True)
                t.start()
            elif kind == Message.EXIT:
                return DecodeEndReason.EXITED
            elif kind == Message.END_EARLY:
                return DecodeEndReason.ENDED_EARLY
            elif kind == Message.BUFFER:
                filepath = body.decode()
                localpath = from_remote_path(pid, filepath)
                ts.trace_buffers.append(tracebuffer.create(localpath))
                os.unlink(localpath)
            elif kind == Message.HEARTBEAT:
                decode_trace_buffers_once(ts.trace_buffers)
            elif kind == Message.AGG_BUFFER:
                filepath = body.decode()
                localpath = from_remote_path(pid, filepath)
                ts.agg_buffers.append(tracebuffer.open_agg_buffer(localpath))
                os.unlink(localpath)
            else:
                print('unknown message kind:', kind, file=sys.stderr)
    finally:
        # But maybe we need to kill it...
        if t is not None and t.ident:
            try:
                signal.pthread_kill(t.ident, signal.SIGINT)
            except ProcessLookupError:
                pass  # It may have finished.
            t.join()


def decode_and_print_remaining(pid: int, s: socket.socket, ts: TraceState):
    # This should not block as the client should have disconnected
    decode_and_print_forever(pid, s, ts, only_print=True)
    decode_trace_buffers_once(ts.trace_buffers)

    for agg_buffer in ts.agg_buffers:
        accumulate_buffer(agg_buffer, agg_buffer.epoch - 1, ts)
        accumulate_buffer(agg_buffer, agg_buffer.epoch, ts)
    print_maps(ts)


def print_maps(ts: TraceState):
    for name, mapp in ts.chkptd_aggs.items():
        try:
            print_map(name, mapp)
        except Exception:
            print(f"Failed to print map: {name}:", file=sys.stderr)
            traceback.print_exc(file=sys.stderr)


def print_map(name, mapp, out=sys.stdout):

    print(name, "\n", file=out)
    kwidth, vwidth = 0, 0
    for k, v in mapp.items():
        kwidth = max(kwidth, len(str(k)))
        vwidth = max(vwidth, len(str(v)))
    for k, v in sorted(mapp.items(), key=lambda kv: kv[1]):
        if isinstance(k, (int, str)):
            key_part = f"  {k:{kwidth}}:"
        else:
            key_part = f"  {k!s:{kwidth}}:"
        if isinstance(v, Quantization):
            value_part = f"\n{v}"
        else:
            value_part = f" {v:{vwidth}}"
        print(f"{key_part}{value_part}", file=out)


@dataclass
class QuantizationRow:
    value: int
    distribution: float
    count: int


def format_quantization(self: Quantization):
    q = self
    # We shamelessly attempt to mimic dtrace here
    total_count = sum([count for count in q.buckets])
    if total_count == 0:
        raise ValueError('empty quantization')

    first_to_print = min(
        (bucket - 1) for bucket, count in enumerate(q.buckets)
        if count > 0
    )
    first_to_print = max(first_to_print, 0)

    last_to_print = max(
        (bucket + 1) for bucket, count in enumerate(q.buckets)
        if count > 0
    )
    last_to_print = min(last_to_print, len(q.buckets) - 1)

    rows: list[QuantizationRow] = []
    for bucket, count in enumerate(q.buckets):
        if first_to_print <= bucket <= last_to_print:
            rows.append(
                QuantizationRow(
                    value=q.idx_value(bucket),  # todo,
                    distribution=(count / total_count),
                    count=count,
                )
            )

    lines = [
        "               value  ------------- Distribution ------------- count"
    ]
    for row in rows:
        parts = []
        parts.append(f"{row.value:20} |")
        parts.append('@' * round(row.distribution * 40))
        parts.append(' ' * round((1 - row.distribution) * 40))
        parts.append(f" {row.count}")
        lines.append("".join(parts))
    return "\n".join(lines)


# We adapt the Quantization here to avoid adding too much tracer related
# code into the tracee
Quantization.__str__ = format_quantization



"""
The module that is imported in the tracee.

This is a single large file to simplify injecting into the target (tracee).
All imports are from the standard library.
"""
import array
import atexit
import inspect
import io
import os
import pickle
import re
import socket
import struct
import sys
import textwrap
import threading
import time
import traceback
from collections import abc, namedtuple
from types import CodeType, FrameType, SimpleNamespace
from typing import Any, Callable, Literal, NoReturn, Optional, Sequence, Union

from pymontrace import tracebuffer

TOOL_ID = sys.monitoring.DEBUGGER_ID if sys.version_info >= (3, 12) else 0


# Replace with typing.assert_never after 3.11
def assert_never(arg: NoReturn) -> NoReturn:
    raise AssertionError(f"assert_never: got {arg!r}")


def glob2re(glob_pattern: str):
    return re.compile('^' + re.escape(glob_pattern).replace('\\*', '.*') + '$')


class InvalidProbe:
    def __init__(self) -> None:
        raise IndexError('Invalid probe ID')


class LineProbe:
    def __init__(self, path: str, lineno: str) -> None:
        if path == '':
            path = '*'
        self.path = path
        self.lineno = int(lineno)

        self.abs = os.path.isabs(path)

        star_count = sum(map(lambda c: c == '*', path))
        self.is_path_endswith = path.startswith('*') and star_count == 1
        self.pathend = path
        if self.is_path_endswith:
            self.pathend = path[1:]
        # TODO: more glob optimizations

        self.isregex = False
        if star_count > 0 and not self.is_path_endswith:
            self.isregex = True
            self.regex = glob2re(path)

    def matches(self, co_filename: str, line_number: int):
        if line_number != self.lineno:
            return False
        return self.matches_file(co_filename)

    def matches_file(self, co_filename: str):
        if self.is_path_endswith:
            return co_filename.endswith(self.pathend)
        if self.abs:
            to_match = co_filename
        else:
            to_match = os.path.relpath(co_filename)
        if self.isregex:
            return bool(self.regex.match(to_match))
        return to_match == self.path

    def __eq__(self, value: object, /) -> bool:
        # Just implemented to help with tests
        if isinstance(value, LineProbe):
            return value.path == self.path and value.lineno == self.lineno
        return False

    @staticmethod
    def listsites(path='*', lineno='*'):
        if path == '':
            path = '*'
        if lineno == '':
            lineno = '*'
        # TODO: allow rangees for lineno
        curdir = os.path.abspath('.')
        probe = LineProbe(path, '0')
        for module in sys.modules.values():
            try:
                filepath = inspect.getfile(module)
            except (OSError, TypeError):
                continue
            if not probe.matches_file(filepath):
                continue
            if filepath.startswith(curdir):
                filepath = os.path.relpath(filepath)
            try:
                lines, start = inspect.getsourcelines(module)
                if start == 0:  # This might always be the case for modules
                    start = 1
                for i, line in enumerate(lines):
                    line = line.removesuffix('\n')
                    this_lineno = i + start
                    if lineno == '*' or str(this_lineno) == lineno:
                        yield f'line:{filepath}:{i + start} {line}'
            except OSError:
                pass


class PymontraceProbe:
    def __init__(self, _: str, hook: Literal['BEGIN', 'END']) -> None:
        self.is_begin = hook == 'BEGIN'
        self.is_end = hook == 'END'

    @staticmethod
    def listsites(unused='', hook='*'):
        if hook == '':
            hook = '*'
        if unused != '':
            return
        pattern = glob2re(hook)
        for h in ('BEGIN', 'END'):
            if pattern.match(h):
                yield f'pymontrace::{h}'


_FUNC_PROBE_EVENT = Literal['start', 'yield', 'resume', 'return', 'unwind']


class FuncProbe:

    # Grouped by the shape of the sys.monitoring callback
    entry_sites = ('start', 'resume')
    return_sites = ('yield', 'return')
    unwind_sites = ('unwind',)

    def __init__(self, qpath: str, site: _FUNC_PROBE_EVENT) -> None:
        if qpath == '':
            qpath = '*'
        for c in qpath:
            if not (c.isalnum() or c in '*._'):
                raise ValueError('invalid qpath glob: {qpath!r}')
        self.qpath = qpath
        self.site = site

        self.name = ""
        self.is_name_match = False
        self.is_star_match = False
        self.is_suffix_path = False
        self.suffix = ""
        self.isregex = False

        star_count = sum(map(lambda c: c == '*', qpath))
        dot_count = sum(map(lambda c: c == '.', qpath))

        if qpath == '*':
            self.is_star_match = True

        # Example: *.foo
        elif qpath.startswith('*.') and star_count == 1 and dot_count == 1:
            self.is_name_match = True
            self.name = qpath[2:]

        # Example: *.bar.foo
        elif qpath.startswith('*.') and star_count == 1 and dot_count > 1:
            self.is_suffix_path = True
            self.suffix = qpath[2:]
            self.name = self.suffix.split('.')[-1]

        elif star_count > 0:
            self.isregex = True
            self.regex = glob2re(qpath)

    def __repr__(self):
        return f'FuncProbe(qpath={self.qpath!r}, site={self.site!r})'

    def excludes(self, code: CodeType) -> bool:
        """fast path for when we don't have the frame yet"""
        if self.is_star_match:
            return False
        if self.is_name_match:
            return code.co_name != self.name
        return False

    def matches(self, frame: FrameType) -> bool:
        if self.is_star_match:
            return True
        if self.is_name_match:
            return frame.f_code.co_name == self.name

        if '__name__' not in frame.f_globals:
            # can happen if an eval/exec gets traced
            module_name = ''
            # It would be interesting to know if there are cases where
            # __name__ is not there but inspect.getmodulename can deduce
            # the name...
        else:
            module_name = frame.f_globals['__name__']

        co_name = frame.f_code.co_name

        if sys.version_info >= (3, 11):
            co_qualname = frame.f_code.co_qualname
            qpath = '.'.join(filter(bool, [module_name, co_qualname]))

            if self.is_suffix_path:
                if co_name != self.name:
                    return False
                return qpath.endswith(self.suffix)
        else:
            if self.is_suffix_path:
                if co_name != self.name:
                    return False
            # This is expensive, that's why we've split the is_suffix_path
            # condition into two parts.
            co_qualname = Frame.get_qualname(frame)
            qpath = '.'.join(filter(bool, [module_name, co_qualname]))
            if self.is_suffix_path:
                # Unsure if this should actually be a return if match
                return qpath.endswith(self.suffix)

        if self.isregex:
            return bool(self.regex.match(qpath))

        # make it simpler to trace simple scripts:
        if frame.f_globals.get('__name__') == '__main__':
            if self.qpath == co_qualname:
                return True

        return self.qpath == qpath

    @staticmethod
    def listsites(qpath='*', site='*'):
        if qpath == '':
            qpath = '*'
        if site == '':
            site = '*'
        funcsites = ('start', 'return', 'unwind')
        if sys.version_info < (3, 12):
            gensites = ()
        else:
            gensites = ('yield', 'resume')

        # it's a bit complicated to make fake frame objects and check matches
        # and we don't refactor to keep the real match code fast

        qp_regex = glob2re(qpath)
        if site != '*':
            site_regex = glob2re(site)
            funcsites = tuple(site for site in funcsites if site_regex.match(site))
            gensites = tuple(site for site in gensites if site_regex.match(site))

        for name, module in sys.modules.copy().items():
            for x, xvalue in vars(module).items():
                if inspect.isfunction(xvalue):
                    if inspect.isfunction(xvalue):
                        if qp_regex.match(qpath := f'{name}.{x}'):
                            for site in funcsites:
                                yield f"func:{qpath}:{site}"
                    if inspect.iscoroutinefunction(xvalue) \
                            or inspect.isgeneratorfunction(xvalue):
                        if qp_regex.match(qpath := f'{name}.{x}'):
                            for site in gensites:
                                yield f"func:{qpath}:{site}"
                elif inspect.isclass(o := xvalue):
                    # We use dir instead of vars to see base class methods
                    # on their subclasses too
                    for y in dir(o):
                        try:
                            field = getattr(o, y, None)
                        except Exception:
                            continue  # some descriptors try to import stuff
                        if field is None:
                            continue
                        if inspect.isfunction(field):
                            if qp_regex.match(qpath := f'{name}.{x}.{y}'):
                                for site in funcsites:
                                    yield f"func:{qpath}:{site}"
                        if inspect.iscoroutinefunction(field) \
                                or inspect.isgeneratorfunction(field):
                            if qp_regex.match(qpath := f'{name}.{x}.{y}'):
                                for site in gensites:
                                    yield f"func:{qpath}:{site}"


class Frame:

    # 3.9 and 3.10 only
    @staticmethod
    def get_qualname(frame: FrameType):
        co_name = frame.f_code.co_name
        if 'self' in frame.f_locals:
            classname = frame.f_locals['self'].__class__.__qualname__
            co_qualname = f"{classname}.{co_name}"
            return co_qualname

        # Is/was it a locally defined function?
        if (parent := frame.f_back) is not None:
            if (func := parent.f_locals.get(co_name)) is not None and \
                    inspect.isfunction(func):
                if frame.f_code is func.__code__:
                    return func.__qualname__
            if (func := parent.f_globals.get(co_name)) is not None and \
                    inspect.isfunction(func):
                if frame.f_code is func.__code__:
                    return func.__qualname__

            for v in parent.f_locals.values():
                if inspect.isfunction(func := v) and \
                        frame.f_code is func.__code__:
                    return func.__qualname__
            # There is another case where it might be a renamed
            # import but this is starting to get rather desperate
        # Fallback
        return co_name


ProbeDescriptor = namedtuple('ProbeDescriptor', ('id', 'name', 'construtor'))

PROBES = {
    0: ProbeDescriptor(0, 'invalid', InvalidProbe),
    1: ProbeDescriptor(1, 'line', LineProbe),
    2: ProbeDescriptor(2, 'pymontrace', PymontraceProbe),
    3: ProbeDescriptor(3, 'func', FuncProbe),
}
PROBES_BY_NAME = {
    descriptor.name: descriptor for descriptor in PROBES.values()
}
ValidProbe = Union[LineProbe, PymontraceProbe, FuncProbe]


def decode_pymontrace_program(encoded: bytes):

    def read_null_terminated_str(buf: bytes) -> tuple[str, bytes]:
        c = 0
        while buf[c] != 0:
            c += 1
        return buf[:c].decode(), buf[c + 1:]

    version, = struct.unpack_from('=H', encoded, offset=0)
    if version != 1:
        # PEF: Pymontrace Encoding Format
        raise ValueError(f'Unexpected PEF version: {version}')
    num_probes, = struct.unpack_from('=H', encoded, offset=2)

    probe_actions: list[tuple[ValidProbe, str]] = []
    remaining = encoded[4:]
    for _ in range(num_probes):
        probe_id, num_args = struct.unpack_from('=BB', remaining)
        remaining = remaining[2:]
        args = []
        for _ in range(num_args):
            arg, remaining = read_null_terminated_str(remaining)
            args.append(arg)
        action, remaining = read_null_terminated_str(remaining)

        ctor = PROBES[probe_id].construtor
        probe_actions.append(
            (ctor(*args), action)
        )
    assert len(remaining) == 0
    return probe_actions


class Message:
    PRINT = 1
    ERROR = 2
    THREADS = 3     # Additional threads the tracer must attach to
    EXIT = 4        # The tracee is exiting (atexit)
    END_EARLY = 5   # The tracing code called the pmt.exit function
    # MAPS = 6        # A new map file is available (TODO: clean up)
    BUFFER = 7      # A new trace buffer has been created
    HEARTBEAT = 8   # Sent to detect tracer death
    AGG_BUFFER = 9  # A new aggregation buffer has been created


class TracerRemote:

    comm_fh: Union[socket.socket, None] = None
    primary_buffer: Union[tracebuffer.TraceBuffer, None] = None

    def __init__(self) -> None:
        self._lock = threading.RLock()
        self.last_heartbeat = time.monotonic()
        self._agg_buffer_count = 0

    @property
    def is_connected(self):
        # Not sure the lock is actually needed here if the GIL is still about
        with self._lock:
            return self.comm_fh is not None

    def connect(self, comm_file: str):
        if self.comm_fh is not None:
            # Maybe a previous settrace failed half-way through
            try:
                self.comm_fh.close()
            except Exception:
                pass
        # TODO: once we don't send data anymore, switch to SOCK_DGRAM
        self.comm_fh = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.comm_fh.connect(comm_file)

    def close(self):
        with self._lock:
            self._close()

    def _close(self):
        try:
            self.comm_fh.close()  # type: ignore  # we catch the exception
        except Exception:
            pass
        self.comm_fh = None
        if self.primary_buffer is not None:
            self.primary_buffer.close()
            self.primary_buffer = None

    def sendall(self, data):
        # Probes may be installed in multiple threads. We lock to avoid
        # mixing messages from different threads onto the socket.
        with self._lock:
            if self.comm_fh is not None:
                try:
                    return self.comm_fh.sendall(data)
                except BrokenPipeError:
                    self._force_close()

    def _force_close(self):
        unsettrace()  # TODO: skip end actions... maybe
        self.close()

    def _heartbeat(self):
        if time.monotonic() - self.last_heartbeat > 1.0:
            self.send_heartbeat()
            self.last_heartbeat = time.monotonic()

    @staticmethod
    def _create_primary_buffer(comm_fh: socket.socket):
        # Should there be a separate one of these for each thread/cpu ?
        fn: str = comm_fh.getpeername() + '.buffer'
        assert fn.startswith('/'), f"{fn=}"
        return fn, tracebuffer.create(fn)

    def writeall(self, data):
        self._heartbeat()

        to_send = []
        with self._lock:
            if self.comm_fh is None:
                return
            if self.primary_buffer is None:
                fn, self.primary_buffer = self._create_primary_buffer(
                    self.comm_fh,
                )
                to_send.append(self._encode_buffer(fn))
            tb = self.primary_buffer
            tb.write(data)
            for msg in to_send:
                self.sendall(msg)

    def create_agg_buffer(self, name: str):
        if (comm_fh := self.comm_fh) is None:
            # can happen if disconnect happens during trace action execution
            raise RuntimeError('no remote')
        with self._lock:
            i = self._agg_buffer_count
            self._agg_buffer_count += 1
        fn: str = comm_fh.getpeername() + f'.map{i:03}'
        assert fn.startswith('/'), f"{fn=}"
        return tracebuffer.create_agg_buffer(name, fn)

    @staticmethod
    def _encode_print(*args, **kwargs):
        message_type = Message.PRINT
        if kwargs.get('file') == sys.stderr:
            message_type = Message.ERROR

        buf = io.StringIO()
        kwargs['file'] = buf
        print(*args, **kwargs)

        to_write = buf.getvalue().encode()
        return struct.pack('=HH', message_type, len(to_write)) + to_write

    @staticmethod
    def _encode_threads(tids):
        count = len(tids)
        fmt = '=HH' + (count * 'Q')
        body_size = struct.calcsize((count * 'Q'))
        return struct.pack(fmt, Message.THREADS, body_size, *tids)

    def notify_threads(self, tids):
        """
        Notify the tracer about additional threads that may need a
        settrace call.
        """
        to_write = self._encode_threads(tids)
        self.sendall(to_write)

    def _send_msg_no_body(self, msg: int):
        body_size = 0
        self.sendall(struct.pack('=HH', msg, body_size))

    def notify_exit(self):
        self._send_msg_no_body(Message.EXIT)

    def notify_end_early(self):
        self._send_msg_no_body(Message.END_EARLY)

    @staticmethod
    def _encode_buffer(filepath: str):
        to_write = filepath.encode()
        body_size = len(to_write)
        return struct.pack('=HH', Message.BUFFER, body_size) + to_write

    def send_heartbeat(self):
        self._send_msg_no_body(Message.HEARTBEAT)

    def notify_agg_buffer(self, filepath: str):
        to_write = filepath.encode()
        body_size = len(to_write)
        self.sendall(struct.pack('=HH', Message.AGG_BUFFER, body_size) + to_write)


remote = TracerRemote()


class PMTError(Exception):
    """Represents a mistake in the use of pmt."""
    pass


class EvaluationError(PMTError):
    pass


class AggOp:
    """Aggregation Operation"""
    NONE = 0
    COUNT = 1
    SUM = 2
    MAX = 3
    MIN = 4
    QUANTIZE = 5
    # How will this work if we add lquantize ??


class aggregation:
    class Base:
        op = AggOp.NONE

        def aggregate(self, current):
            raise NotImplementedError

    class Count(Base):
        op = AggOp.COUNT

        def aggregate(self, current: Optional[int]):
            return (current or 0) + 1

    class Sum(Base):
        op = AggOp.SUM

        def __init__(self, value):
            self.value = value

        def aggregate(self, current):
            return (current or 0) + self.value

    class Max(Base):
        op = AggOp.MAX

        def __init__(self, value):
            self.value = value

        def aggregate(self, current):
            return self.value if current is None else max(current, self.value)

    class Min(Base):
        op = AggOp.MIN

        def __init__(self, value):
            self.value = value

        def aggregate(self, current):
            return self.value if current is None else min(current, self.value)

    class Quantize(Base):
        op = AggOp.QUANTIZE

        def __init__(self, value):
            self.value = value

        def aggregate(self, current: Optional['Quantization']):
            quant = current or Quantization()
            quant.add(self.value)
            return quant


class agg:
    """
    A namespace of functions to be used in the actions.

    e.g. 'line:xxx.py:123 {{ maps[ctx.route] = agg.quantize(ctx.duration) }}'
    """
    @staticmethod
    def count():
        return aggregation.Count()

    @staticmethod
    def sum(value):
        return aggregation.Sum(value)

    @staticmethod
    def min(value):
        return aggregation.Min(value)

    @staticmethod
    def max(value):
        return aggregation.Max(value)

    @staticmethod
    def quantize(value):
        return aggregation.Quantize(value)


class Quantization:

    zero_idx = 64

    def __init__(self) -> None:
        self.buckets = array.array('Q', [0] * 128)

    # used in accumulation in tracer
    def __add__(self, other):
        if not isinstance(other, Quantization):
            raise TypeError
        result = Quantization()
        for i, (v0, v1) in enumerate(zip(self.buckets, other.buckets)):
            result.buckets[i] = v0 + v1
        return result

    def add(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError(f"quantize: not int or float: {value!r}")
        # The highest power of two less than or equal to value

        value = int(value)
        bucket_idx = self.bucket_idx(value)
        if bucket_idx >= len(self.buckets):
            # perhaps introduce some kind of "scale" property ?
            raise ValueError('large number not yet quantizable: {value!r}')
        self.buckets[bucket_idx] += 1

    @classmethod
    def bucket_idx(cls, value: int) -> int:
        if value >= 0:
            return cls.zero_idx + value.bit_length()
        else:
            return cls.zero_idx - value.bit_length()

    @classmethod
    def idx_value(cls, bucket_idx: int) -> int:
        if bucket_idx == cls.zero_idx:
            return 0
        if bucket_idx < cls.zero_idx:
            return - 2**(-(1 + bucket_idx - cls.zero_idx))
        else:
            return 2**(bucket_idx - 1 - cls.zero_idx)


class VarNS(SimpleNamespace):

    def __setattr__(self, name: str, value, /) -> None:

        current = getattr(self, name, None)
        if isinstance(value, aggregation.Base):
            new = value.aggregate(current)
            object.__setattr__(self, name, new)
        else:
            object.__setattr__(self, name, value)


class PMTMap(abc.MutableMapping):
    def __init__(self, buffer: tracebuffer.AggBuffer) -> None:
        super().__init__()
        self.index: dict[Any, tuple[int, int]] = {}
        self.buffer = buffer
        self.agg_op = 0
        self.epoch = 2

    def __len__(self) -> int:
        return len(self.index)

    def __iter__(self):
        return iter(self.index)

    def _reset_index(self):
        self.epoch = self.buffer.epoch
        self.index = {}

    def __getitem__(self, key, /):
        with self.buffer:
            if self.epoch != self.buffer.epoch:
                self._reset_index()
            value, _ = self._getitem(key)
            return value

    _NO_DEFAULT = object()

    def _getitem(self, key, default=_NO_DEFAULT):
        try:
            offset, size = self.index[key]
        except KeyError:
            if default != self._NO_DEFAULT:
                return default, None
            raise
        data = self.buffer.read(offset, size)
        value = tracebuffer.decode_value(data, Quantization)
        return value, data

    @staticmethod
    def _encode(key, value) -> bytes:
        # pickle does not encode with a consistent size for same type
        # so we cannot use it for the value, which changes
        key_data = pickle.dumps(key)

        return (len(key_data).to_bytes(4, sys.byteorder)
                + key_data
                + PMTMap._encode_value(value))

    @staticmethod
    def _encode_value(value) -> bytes:
        value_data: bytes
        if isinstance(value, int):
            if value < 0:
                # we use `=` (standard) instead of '@' (native)
                # because we are anyway out of alignment
                value_data = b'q' + struct.pack('=q', value)
            else:
                value_data = b'Q' + struct.pack('=Q', value)
        elif isinstance(value, float):
            value_data = b'd' + struct.pack('=d', value)
        elif isinstance(value, Quantization):
            value_data = b'Y' + value.buckets.tobytes()
        else:
            raise TypeError(
                f'unsupported aggregation value type: {value.__class__.__name__}'
            )
        return (len(value_data).to_bytes(4, sys.byteorder)
                + value_data)

    @staticmethod
    def _decode(data: bytes) -> tuple:
        key_length = int.from_bytes(data[:4], byteorder=sys.byteorder)
        assert key_length <= len(data) - 4, f"{key_length} > {len(data[4:])}"

        key = pickle.loads(data[4:])  # ignores trailing
        value_length = int.from_bytes(data[4 + key_length:][:4], sys.byteorder)
        assert 4 + key_length + 4 + value_length == len(data), \
            f"4 + {key_length} + 4 + {value_length} != {len(data)}"

        value = tracebuffer.decode_value(data, Quantization)
        return key, value

    @staticmethod
    def _decode_value(data: bytes) -> Union[int, float, Quantization]:
        key_length = int.from_bytes(data[:4], sys.byteorder)
        value_data = data[8 + key_length:]
        if (prefix := value_data[:1]) in (b'qQd'):
            (value,) = struct.unpack('=' + prefix.decode(), value_data[1:])
        elif prefix == b'Y':
            value = object.__new__(Quantization)
            value.buckets = array.array('Q')
            value.buckets.frombytes(value_data[1:])
        else:
            raise ValueError(f'bad aggregation value with prefix {prefix}')
        return value

    def _setitem(self, key, value, old_data):
        if key in self.index:
            offset, size = self.index[key]
            new_value_data = tracebuffer.encode_value(value, Quantization)
            data = old_data[:-len(new_value_data)] + new_value_data
            self.buffer.update(data, offset, size)
        else:
            data = tracebuffer.encode_entry(key, value, Quantization)
            offset, size = self.buffer.write(data)
            self.index[key] = (offset, size)

    def __setitem__(self, key, value, /) -> None:
        with self.buffer:
            if self.epoch != self.buffer.epoch:
                self._reset_index()
            current, old_data = self._getitem(key, None)
            if isinstance(value, aggregation.Base):
                new = value.aggregate(current)
                self._setitem(key, new, old_data)
                if self.agg_op == 0:
                    self.buffer.agg_op = value.op
                    self.agg_op = value.op
            elif isinstance(value, (int, float, Quantization)):
                self._setitem(key, value, old_data)
            else:
                raise EvaluationError(
                    f'unsupported value type: {value.__class__.__name__}, '
                    'use an aggregation function such as agg.count() instead'
                )

    def __delitem__(self, key):
        # Maybe we could reset the value to 0
        raise NotImplementedError


class MapNS(SimpleNamespace):

    # Only happens on AttributeError
    def __getattr__(self, name: str):
        if name.startswith("__") or name == '_is_coroutine_marker':
            raise AttributeError(f"'MapNS' object has no attribute {name!r}")
        # At this point we create the buffer space
        buffer = remote.create_agg_buffer(name)
        new_map = PMTMap(buffer)
        self.__setattr__(name, new_map)
        remote.notify_agg_buffer(buffer.fn)
        return new_map

    def __setattr__(self, name: str, value, /) -> None:
        if not isinstance(value, PMTMap):
            # ... The user is probably messing up some syntax..
            raise EvaluationError(
                f'cannot set attribute {name!r} on maps, '
                f'use maps[...] = ... or vars.{name} instead'
            )
        super().__setattr__(name, value)

    def __getitem__(self, key):
        # A default map. We use a name that's not possible in python
        # to avoid possible conflicts
        return getattr(self, '@')[key]

    def __setitem__(self, key, value, /) -> None:
        return getattr(self, '@').__setitem__(key, value)

    def __delitem__(self, key):
        return getattr(self, '@').__delitem__(key)


class pmt:
    """
    pmt is a utility namespace of functions that may be useful for examining
    the system and returning data to the tracer.
    """

    @staticmethod
    def print(*args, **kwargs):
        if remote.is_connected:
            to_write = remote._encode_print(*args, **kwargs)
            remote.writeall(to_write)

    @staticmethod
    def exit(status=None):
        unsettrace(preclose=remote.notify_end_early())

    def __init__(self, frame: Optional[FrameType]):
        self._frame = frame
        self._frozen = True

    def funcname(self):
        if self._frame is None:
            return None
        return self._frame.f_code.co_name

    def qualname(self, Frame=Frame):
        frame = self._frame
        if frame is None:
            return None
        module_name = frame.f_globals.get('__name__')
        if sys.version_info < (3, 11):
            co_qualname = Frame.get_qualname(frame)
        else:
            co_qualname = frame.f_code.co_qualname
        if module_name:
            return f'{module_name}.{co_qualname}'
        return co_qualname

    def args(self):
        frame = self._frame
        if frame is None:
            return None
        f_code = frame.f_code
        num_args = (f_code.co_argcount + f_code.co_kwonlyargcount)
        result = {}
        for name in f_code.co_varnames[:num_args]:
            result[name] = frame.f_locals[name]
        return result

    def printprobes(self, probename, *args):
        # Used to handle the -l flag
        # ... could use PROBES_BY_NAME and/or could glob match the probename
        if probename == 'pymontrace':
            for probesite in PymontraceProbe.listsites(*args):
                pmt.print(probesite)
        elif probename == 'func':
            for probesite in FuncProbe.listsites(*args):
                pmt.print(probesite)
        elif probename == 'line':
            for probesite in LineProbe.listsites(*args):
                pmt.print(probesite)

    _end_actions: list[tuple[PymontraceProbe, CodeType, str]] = []

    vars = VarNS()
    maps = MapNS()

    @staticmethod
    def _reset():
        pmt._end_actions = []
        pmt.vars = VarNS()
        pmt.maps = MapNS()

    def __setattr__(self, name: str, value, /) -> None:
        # It's quite easy too accidentally assign to something here
        # instead of into pmt.vars. This should help.
        if getattr(self, "_frozen", False):
            raise EvaluationError(
                f'cannot set {name!r}, pmt is readonly, '
                'set your attribute on pmt.vars or pmt.maps instead'
            )
        return super().__setattr__(name, value)

    def _asdict(self):
        o = {}
        for k in (vars(self) | vars(pmt)):
            if not k.startswith("_"):
                # seems to be necessary in python 3.9 to access the
                # staticmethods through the instance ... :shrug:
                o[k] = getattr(self, k)
        return o


class ChainNS(SimpleNamespace):
    def __init__(self, level1, level2, /, **kwargs):
        self.__dict__.update(level1)
        self.__dict__.update(kwargs)
        self._level2 = level2

    def __getattr__(self, name: str):
        return getattr(self._level2, name)


def safe_eval(action: CodeType, frame: FrameType, snippet: str):
    try:
        framepmt = pmt(frame)
        globals_as_ns = SimpleNamespace(**frame.f_globals)
        eval(action, {
            'ctx': ChainNS(
                frame.f_locals,
                globals_as_ns,
                # to allow getting around shadowed variables
                globals=globals_as_ns,
            ),
            'pmt': framepmt,
            **framepmt._asdict(),
            'agg': agg,
        }, {})
    except Exception as e:
        _handle_eval_error(e, snippet, frame)


def safe_eval_no_frame(action: CodeType, snippet: str, exit):
    try:
        noframepmt = pmt(frame=None)
        object.__setattr__(noframepmt, 'exit', exit)
        eval(action, {
            'pmt': noframepmt,
            **noframepmt._asdict(),
            'agg': agg,
        }, {})
    except Exception as e:
        _handle_eval_error(e, snippet)


def _handle_eval_error(
    e: Exception, snippet: str, frame: Optional[FrameType] = None,
) -> None:
    buf = io.StringIO()
    print('Probe action failed:', file=buf)
    if isinstance(e, PMTError):
        if frame is not None:
            traceback.print_stack(frame, file=buf)
        print(f"{e.__class__.__name__}: {e}", file=buf)
    else:
        traceback.print_exc(file=buf)
    print(textwrap.indent(snippet, 4 * ' '), file=buf)
    pmt.print(buf.getvalue(), end='', file=sys.stderr)


TraceFunction = Callable[[FrameType, str, Any], Union['TraceFunction', None]]


# Handlers for 3.11 and earlier - TODO: should this be guarded?
def create_event_handlers(
    probe_actions: Sequence[tuple[Union[LineProbe, FuncProbe], CodeType, str]],
):

    if sys.version_info < (3, 10):
        # https://github.com/python/cpython/blob/3.12/Objects/lnotab_notes.txt
        def num_lines(f_code: CodeType):
            lineno = addr = 0
            it = iter(f_code.co_lnotab)
            for addr_incr in it:
                line_incr = next(it)
                addr += addr_incr
                if line_incr >= 0x80:
                    line_incr -= 0x100
                lineno += line_incr
            return lineno
    else:
        def num_lines(f_code: CodeType):
            lineno = f_code.co_firstlineno
            for (start, end, this_lineno) in f_code.co_lines():
                if this_lineno is not None:
                    lineno = max(lineno, this_lineno)
            return lineno - f_code.co_firstlineno

    def make_local_handler(probe, action: CodeType, snippet: str) -> TraceFunction:
        if isinstance(probe, LineProbe):
            def handle_local(frame, event, _arg) -> Optional[TraceFunction]:
                if event != 'line' or probe.lineno != frame.f_lineno:
                    return handle_local
                safe_eval(action, frame, snippet)
                return None

        elif isinstance(probe, FuncProbe) and probe.site == 'return':
            # BUG: Both this event and 'exception' fire during an
            # exception
            def handle_local(frame, event, _arg) -> Optional[TraceFunction]:
                if event == 'return':
                    safe_eval(action, frame, snippet)
                    return None
                return handle_local
            return handle_local

        elif isinstance(probe, FuncProbe) and probe.site == 'unwind':
            def handle_local(frame, event, _arg) -> Optional[TraceFunction]:
                if event == 'exception':
                    safe_eval(action, frame, snippet)
                    return None
                return handle_local
            return handle_local
        else:
            def handle_local(frame, event, _arg) -> Optional[TraceFunction]:
                return handle_local
            return handle_local
        return handle_local

    def combine_handlers(handlers):
        def handle(frame, event, arg):
            for h in handlers:
                result = h(frame, event, arg)
                if result is None:
                    return None
            return handle
        return handle

    count_line_probes = 0
    count_exit_probes = 0
    count_start_probes = 0
    for (probe, action, snippet) in probe_actions:
        if isinstance(probe, LineProbe):
            count_line_probes += 1
        elif isinstance(probe, FuncProbe) and probe.site == 'start':
            count_start_probes += 1
        elif isinstance(probe, FuncProbe) and probe.site in ('return', 'unwind'):
            count_exit_probes += 1

    # We allow that only one probe will match any given event
    probes_and_handlers = [
        (probe, action, snippet, make_local_handler(probe, action, snippet))
        for (probe, action, snippet) in probe_actions
    ]

    if count_line_probes > 0 and (count_start_probes == 0 and count_exit_probes == 0):
        def handle_call(frame: FrameType, event, arg) -> Union[TraceFunction, None]:
            for probe, action, snippet, local_handler in probes_and_handlers:
                assert isinstance(probe, LineProbe)
                if probe.lineno < frame.f_lineno:
                    continue
                f_code = frame.f_code
                if not probe.matches_file(f_code.co_filename):
                    continue
                if probe.lineno > f_code.co_firstlineno + num_lines(f_code):
                    continue
                return local_handler
            return None
        return handle_call

    if count_line_probes == 0:
        def handle_call(frame: FrameType, event, arg) -> Union[TraceFunction, None]:
            local_handlers = []
            for probe, action, snippet, local_handler in probes_and_handlers:
                assert isinstance(probe, FuncProbe)
                # first just entry
                if probe.site in ('start', 'return', 'unwind') and probe.matches(
                    frame
                ):
                    if probe.site == 'start':
                        safe_eval(action, frame, snippet)
                        continue
                    else:
                        # There are no line probes
                        frame.f_trace_lines = False
                        local_handlers.append(local_handler)
            if len(local_handlers) == 1:
                return local_handlers[0]
            if len(local_handlers) > 1:
                return combine_handlers(local_handlers)
            return None
        return handle_call

    def handle_call(frame: FrameType, event, arg) -> Union[TraceFunction, None]:
        local_handlers = []
        for probe, action, snippet, local_handler in probes_and_handlers:
            if isinstance(probe, LineProbe):
                if probe.lineno < frame.f_lineno:
                    continue
                f_code = frame.f_code
                if not probe.matches_file(f_code.co_filename):
                    continue
                if probe.lineno > f_code.co_firstlineno + num_lines(f_code):
                    continue
                local_handlers.append(local_handler)
            elif isinstance(probe, FuncProbe) and probe.site == 'start':
                if probe.matches(frame):
                    safe_eval(action, frame, snippet)
            elif isinstance(probe, FuncProbe) and probe.site in ('return', 'unwind'):
                if probe.matches(frame):
                    local_handlers.append(local_handler)
        if len(local_handlers) == 1:
            return local_handlers[0]
        if len(local_handlers) > 1:
            return combine_handlers(local_handlers)
        return None

    return handle_call


def connect(comm_file: str):
    """
    Connect back to the tracer.
    Tracer invokes this in the target when attaching to it.
    """
    remote.connect(comm_file)


if sys.version_info >= (3, 12):

    # We enumerate the ones we use so that it's easier to
    # unregister callbacks for them
    class events:
        LINE = sys.monitoring.events.LINE
        PY_START = sys.monitoring.events.PY_START
        PY_RESUME = sys.monitoring.events.PY_RESUME
        PY_YIELD = sys.monitoring.events.PY_YIELD
        PY_RETURN = sys.monitoring.events.PY_RETURN
        PY_UNWIND = sys.monitoring.events.PY_UNWIND

        @classmethod
        def all(cls):
            for k, v in cls.__dict__.items():
                if not k.startswith("_") and isinstance(v, int):
                    yield v


# The function called inside the target to start tracing
def settrace(encoded_program: bytes, is_initial=True):
    try:
        probe_actions = decode_pymontrace_program(encoded_program)

        pmt_probes: list[tuple[PymontraceProbe, CodeType, str]] = []
        line_probes: list[tuple[LineProbe, CodeType, str]] = []
        func_probes: list[tuple[FuncProbe, CodeType, str]] = []

        for probe, user_python_snippet in probe_actions:
            user_python_obj = compile(
                user_python_snippet, '<pymontrace expr>', 'exec'
            )
            # Will support more probes in future.
            assert isinstance(probe, (LineProbe, PymontraceProbe, FuncProbe)), \
                f"Bad probe type: {probe.__class__.__name__}"
            if isinstance(probe, LineProbe):
                line_probes.append((probe, user_python_obj, user_python_snippet))
            elif isinstance(probe, PymontraceProbe):
                pmt_probes.append((probe, user_python_obj, user_python_snippet))
            elif isinstance(probe, FuncProbe):
                func_probes.append((probe, user_python_obj, user_python_snippet))
            else:
                assert_never(probe)

        pmt._end_actions = [
            (probe, action, snippet)
            for (probe, action, snippet) in pmt_probes
            if probe.is_end
        ]
        for (probe, action, snippet) in pmt_probes:
            if probe.is_begin:
                safe_eval_no_frame(action, snippet, exit=early_exit)

        if sys.version_info < (3, 12):
            # TODO: handle func probes
            event_handlers = create_event_handlers(line_probes + func_probes)
            sys.settrace(event_handlers)
            if is_initial:
                threading.settrace(event_handlers)
                own_tid = threading.get_native_id()
                additional_tids = [
                    thread.native_id for thread in threading.enumerate()
                    if (thread.native_id != own_tid
                        and thread.native_id is not None)
                ]
                if additional_tids:
                    remote.notify_threads(additional_tids)
        else:

            def handle_line(code: CodeType, line_number: int):
                for (probe, action, snippet) in line_probes:
                    if not probe.matches(code.co_filename, line_number):
                        continue
                    if ((cur_frame := inspect.currentframe()) is None
                            or (frame := cur_frame.f_back) is None):
                        # TODO: warn about not being able to collect data
                        continue
                    safe_eval(action, frame, snippet)
                    return None
                return sys.monitoring.DISABLE

            start_probes = [p for p in func_probes if p[0].site == 'start']
            resume_probes = [p for p in func_probes if p[0].site == 'resume']
            yield_probes = [p for p in func_probes if p[0].site == 'yield']
            return_probes = [p for p in func_probes if p[0].site == 'return']
            unwind_probes = [p for p in func_probes if p[0].site == 'unwind']

            # For any func probe except unwind
            def handle_(
                probes: list[tuple[FuncProbe, CodeType, str]],
                nodisable=False,
            ):
                def handle(code: CodeType, arg1, arg2=None):
                    for (probe, action, snippet) in probes:
                        if probe.excludes(code):
                            continue
                        if ((cur_frame := inspect.currentframe()) is None
                                or (frame := cur_frame.f_back) is None):
                            continue
                        if not probe.matches(frame):
                            continue
                        safe_eval(action, frame, snippet)
                        return None
                    if nodisable:
                        return None
                    return sys.monitoring.DISABLE
                return handle

            sys.monitoring.use_tool_id(TOOL_ID, 'pymontrace')

            event_set: int = 0
            handlers = [
                (events.LINE, line_probes, handle_line),
                (events.PY_START, start_probes, handle_(start_probes)),
                (events.PY_RESUME, resume_probes, handle_(resume_probes)),
                (events.PY_YIELD, yield_probes, handle_(yield_probes)),
                (events.PY_RETURN, return_probes, handle_(return_probes)),
                (events.PY_UNWIND, unwind_probes, handle_(unwind_probes, nodisable=True)),
            ]
            for event, probes, handler in handlers:
                if len(probes) > 0:
                    sys.monitoring.register_callback(
                        TOOL_ID, event, handler
                    )
                    event_set |= event

            sys.monitoring.set_events(TOOL_ID, event_set)

        atexit.register(exithook)

    except Exception as e:
        try:
            buf = io.StringIO()
            print(f'{__name__}.settrace failed', file=buf)
            traceback.print_exc(file=buf)
            pmt.print(buf.getvalue(), end='', file=sys.stderr)
        except Exception:
            print(f'{__name__}.settrace failed:', repr(e), file=sys.stderr)
        remote.close()


def synctrace():
    """
    Called in each additional thread by the tracer.
    """
    # sys.settrace must be called in each thread that wants tracing
    if sys.version_info < (3, 10):
        sys.settrace(threading._trace_hook)  # type: ignore  # we're adults
    elif sys.version_info < (3, 12):
        sys.settrace(threading.gettrace())
    else:
        pass  # sys.monitoring should already have all threads covered.


def exithook():
    unsettrace(preclose=remote.notify_exit)


def unsettrace(preclose=None):
    atexit.unregister(exithook)
    # This can fail if installing probes failed.
    try:
        if sys.version_info < (3, 12):
            threading.settrace(None)  # type: ignore  # bug in typeshed.
            sys.settrace(None)
        else:
            for event in events.all():
                sys.monitoring.register_callback(
                    TOOL_ID, event, None
                )
            sys.monitoring.set_events(
                TOOL_ID, sys.monitoring.events.NO_EVENTS
            )
            sys.monitoring.free_tool_id(TOOL_ID)

        for (probe, action, snippet) in pmt._end_actions:
            assert probe.is_end
            safe_eval_no_frame(action, snippet, exit=null_exit)

        pmt._reset()
        if preclose is not None:
            preclose()
        remote.close()
    except Exception:
        print(f'{__name__}.unsettrace failed', file=sys.stderr)
        traceback.print_exc(file=sys.stderr)


# An exit that can be called during BEGIN probes, before tracing has been
# installed.
def early_exit(status=None):
    remote.notify_end_early()
    remote.close()


# An exit that can be called during END probes. We ignore exit in END probes
def null_exit(status=None):
    pass



import os
import sys
from setuptools import Extension, setup

DARWIN_SOURCES = [
    "c_src/darwin_64bit.c",
    "c_src/mach_excServer.c",
]

LINUX_SOURCES = [
    "c_src/attacher_linux_64bit.c",
]

sources = [
    "c_src/attachermodule.c",
]
if sys.platform == 'darwin':
    sources += DARWIN_SOURCES
elif sys.platform == 'linux':
    sources += LINUX_SOURCES
else:
    print(sys.platform, 'is not currently supported...', file=sys.stderr)

# Undefining NDEBUG enables some debug logging.
undef_macros = []
if os.getenv("PYMONTRACE_DEBUG") in ("1", "true", "True", "yes"):
    undef_macros = ["NDEBUG"]

setup(
    ext_modules=[
        Extension(
            name="pymontrace.attacher",
            sources=sources,
            undef_macros=undef_macros,
            py_limited_api=True,
        ),
        Extension(
            name="pymontrace._tracebuffer",
            sources=["c_src/_tracebuffermodule.c"],
            undef_macros=undef_macros,
            py_limited_api=True,
        ),
    ],
    # https://github.com/pypa/setuptools/issues/4741
    options={'bdist_wheel': {'py_limited_api': 'cp39'}},
)
